package Uebung12;

import java.util.Scanner;

class WrongProbabilityException extends RuntimeException {
    WrongProbabilityException(String msg) {
        super(msg);
    }
}

/**
 * Repräsentiert ein Spielfeld des game of life.
 */
class Life {
    Stone[][] board;
    double m_probabilityNever = 0.05;
    double m_probabilityAlways = 0.05;

    /**
     * Erstellt ein neues Life-Objekt
     * @param zeilen Anzahl der Zeilen
     * @param spalten Anzahl der Spalten
     */
    public Life(int zeilen, int spalten) {
        this.board = new Stone[zeilen][spalten];
        initializeArray();
    }

    /**
     * Erstellt ein neues Life-Objekt
     * @param zeilen Anzahl der Zeilen
     * @param spalten Anzahl der Spalten
     * @param probabilityNever Wahrscheinlichkeit, dass Zelle ein NeverStone ist
     *                         (probabilityNever + probabilityAlways <= 1.0)
     * @param probabilityAlways Wahrscheinlichkeit, dass Zelle ein AlwaysStone ist
     *                          (probabilityNever + probabilityAlways <= 1.0)
     */
    public Life(int zeilen, int spalten, double probabilityNever, double probabilityAlways) {
        this(zeilen, spalten);
        if (probabilityAlways + probabilityNever > 1.0) {
            throw new WrongProbabilityException("Die Wahrscheinlichkeit für AlwaysStone und NeverStone ist größer 1!");
        }
        this.m_probabilityNever = probabilityNever;
        this.m_probabilityAlways = probabilityAlways;
    }

    /**
     * Initialisiert ein zweidimensionales-int-Array.
     */
    private void initializeArray() {
        for (int i = 0; i < this.board.length; ++i) {
            for (int j = 0; j < this.board[i].length; ++j) {
                double random = Math.random();
                if (random < m_probabilityNever) {
                    this.board[i][j] = new NeverStone();
                } else if (random < (m_probabilityNever + m_probabilityAlways)) {
                    this.board[i][j] = new AlwaysStone();
                } else {
                    this.board[i][j] = new Stone(Math.random() > 0.5 ? 1 : 0);
                }
            }
        }

        // set Stone neighbours
        final int row = this.board.length;
        for (int zeile = 0; zeile < row; ++zeile) {
            final int column = this.board[zeile].length;
            for (int spalte = 0; spalte < column; ++spalte) {
                int id = 0;
                for (int i = -1; i < 2; ++i) {
                    for (int j = -1; j < 2; ++j) {
                        if (i != 0 || j != 0) {
                            this.board[zeile][spalte].setNeighbour(id, this.board[(zeile + i + row) % row]
                                    [(spalte + j + column) % column]);
                            id++;
                        }
                    }
                }
            }
        }
    }

    /**
     * Berechnet das zweidimensionale-int-Array nach den neuen Vorgaben.
     */
    public void recalculateArray() {
        // Prüft die Steine in der aktuellen Iteration
        for (Stone[] row: this.board) {
            for (Stone cell: row) {
                cell.calculateNextLevel();
            }
        }

        // Speichert die Änderungen der aktuellen Iteration für die Folgende.
        for (Stone[] row: this.board) {
            for (Stone cell: row) {
                cell.nextLevel();
            }
        }
    }

    /**
     * Gibt das übergebene zweidimensionale-boolean-Array zeilenweise auf StOut aus.
     */
    public void printArray() {
        // Array ausgeben
        for (Stone[] zeile : this.board) { // Die Objekte von ausgabe[][] sind zeile[]
            for (Stone zelle : zeile) { // Die Objekte von zeile[] sind zelle
                System.out.print(zelle);
            }
            System.out.println();
        }
    }
}

/**
 * Repräsentiert ein Stein im game of life.
 */
class Stone {
    protected int age;
    protected int nextage;
    protected Stone[] neighbours = new Stone[8];


    /**
     * Erstellt ein neues Stone-Objekt.
     * @param age Alter des Stones.
     */
    public Stone(int age) { this.age = age; }

    protected Stone() { this.age = 0; }

    public void setNeighbour(int id, Stone neighbour) {
        this.neighbours[id] = neighbour;
        /* 0 1 2
           3 X 4
           5 6 7  */
    }

    /**
     * Berechnet die Nachbarschaft des Steins
     * @return berechnete Nachbarschaft nach den neuen Vorgaben.
     */
    protected int getNeighborhood() {
        int nachbarschaft = 0;

        for (Stone neighbour: neighbours) {
            if (neighbour.isAlive()) {
                ++nachbarschaft;
            }
        }

        return nachbarschaft;
    }

    /**
     * Berechnet die nextage-Variable für diesen Stone
     */
    protected void calculateNextLevel() {
        int neighbourhood = this.getNeighborhood();
        if (neighbourhood == 3 || (neighbourhood == 2 && this.isAlive())) {
            this.aging();
        } else {
            this.die();
        }
    }

    /**
     * Gibt zurück, ob der Stone lebt.
     * @return Ob Stone lebendig ist.
     */
    protected boolean isAlive() { return this.age > 0; }

    /**
     * Tötet den Stein.
     */
    protected void die() { this.nextage = 0; }

    /**
     * Lässt den Stein um eine Iteration ältern.
     */
    protected void aging() { this.nextage = age + 1; }

    /**
     * Bereitet Stein auf die nächste Iteration vor.
     */
    protected void nextLevel() { this.age = nextage; }

    /**
     * Gibt das Zeichen, dass das Alter des Stones repräsentiert als Char zurück.
     * @return Zeichen, dass das Alter des Stones repräsentiert.
     */
    @Override
    public String toString() {
        char c = switch (this.age) {
            case 0 -> ' ';
            case 1 -> '.';
            case 2 -> 'o';
            case 3 -> 'O';
            default -> '*';
        };
        return String.valueOf(c);
    }
}

class AlwaysStone extends Stone {

    @Override
    public String toString() {
        return "@";
    }

    public boolean isAlive() { return true; }
}

class NeverStone extends Stone {

    @Override
    public String toString() {
        return "✝";
    }

    public boolean isAlive() { return false; }
}

/**
 * <h1>Übung 12, Aufgabe 01</h1>
 *
 * <p>Programmieren 1
 * Wintersemester 2021/-22
 * HS-Bremerhaven
 * Prof. Dr. Peter Kelb</p>
 *
 * <p>Erweitern Sie Ihr <i>Stone</i> Klasse um ein Array, so dass sich jede Zelle Ihre
 * Nachbarn direkt merkt.</p>
 * <p>Stellen Sie Ihre Methoden auf diese neue Art der Speicherung um.</p>
 * <p>Leiten Sie von der Klasse Stone zwei Klassen <i>AlwaysStone</i> und <i>NeverStone</i> ab.</p>
 * <p>Die <i>AlwaysStone</i> Klasse repräsentiert eine Zelle, die immer besetzt ist, egal, wie
 * die Nachbarschaft aussieht.</p>
 * <p>Die <i>NeverStone</i> Klasse repräsentiert eine Zelle, die niemals besetzt ist, egal, wie
 * die Nachbarschaft aussieht.</p>
 * <p>Verwenden Sie diese beiden neuen Klassen zusätzlich zu der <i>Stone</i> Klasse bei
 * der Initialisierung der <i>Life</i> Klasse.</p>
 *
 * @author franjogiessel
 */
public class Aufgabe01 {
    public static void main(String[] args) {
        Scanner sc = new Scanner(System.in); // neues Scanner-Objekt sc

        System.out.print("Anzahl Zeilen: ");
        int zeilen = sc.nextInt();
        System.out.print("Anzahl Spalten: ");
        int spalten = sc.nextInt();
        System.out.print("Anzahl Tage: ");
        int days = sc.nextInt();

        Life life = new Life(zeilen, spalten);
        life.printArray();

        for (int i = 1; i <= days; ++i) {
            System.out.println("------------------");
            System.out.println("Nach " + i + " Tag" + (i == 1 ? "" : "en" ) + ":");
            life.recalculateArray();
            life.printArray();
        }

    }
}